#include <linux/types.h>
#include <linux/io.h>
#include <stdio.h>

#define PCI_INTERRUPT_LINE	0x3c
#define PCI_VENDOR_ID		0x00
#define ARRAY_SIZE(arr)		(sizeof(arr) / sizeof((arr)[0]))

#define outb(val, reg)	linux_outb(val, reg)
#define outw(val, reg)	linux_outw(val, reg)

#define inb(reg)	linux_inb(reg)
#define inw(reg)	linux_inw(reg)

typedef unsigned long device_t;	// I don't know include which header file

extern u32 _pci_conf_readn(device_t tag, int reg, int width);
extern void _pci_conf_writen(device_t tag, int reg, u32 data, int width);
extern u32 _pci_make_tag(u32 bus, u32 device, u32 function);

static void inline pci_read_config_word(device_t dev, u32 reg, u16 * val)
{
	*val = _pci_conf_read16(dev, reg);
}

static void inline pci_read_config_byte(device_t dev, u32 reg, u8 * val)
{
	*val = _pci_conf_read8(dev, reg);
}

static void inline pci_write_config_word(device_t dev, u32 reg, u16 val)
{
	_pci_conf_writen(dev, reg, val, 2);
}

static void inline pci_write_config_byte(device_t dev, u32 reg, u8 val)
{
	_pci_conf_writen(dev, reg, val, 1);
}

/*
 * SB700 PIC has 10 interrupt input pins, not increasing by number,
 * no 6,7,8 exist.
 */
enum pic_int {
	PIC_INTA = 0,
	PIC_INTB = 1,
	PIC_INTC = 2,
	PIC_INTD = 3,
	PIC_SCI = 4,
	PIC_SMBUS = 5,
	PIC_INTE = 9,
	PIC_INTF = 10,
	PIC_INTG = 11,
	PIC_INTH = 12
};

/*
 * SB700 PIC has 16 interrupt output pins
 */
enum pic_irq {
	PIC_IRQ0 = 0,
	PIC_IRQ1,
	PIC_IRQ2,
	PIC_IRQ3,
	PIC_IRQ4,
	PIC_IRQ5,
	PIC_IRQ6,
	PIC_IRQ7,
	PIC_IRQ8,
	PIC_IRQ9,
	PIC_IRQ10,
	PIC_IRQ11,
	PIC_IRQ12,
	PIC_IRQ13,
	PIC_IRQ14,
	PIC_IRQ15
};

struct pic_routing {
	enum pic_int from;
	enum pic_irq to;
};

struct pic_routing pic_routing_array[] = {
	{PIC_INTA, PIC_IRQ3},	/* 0 */
	{PIC_INTB, PIC_IRQ3},	/* 1 */
	{PIC_INTC, PIC_IRQ6},	/* 2 */
	{PIC_INTD, PIC_IRQ5},	/* 3 */
	{PIC_SCI, PIC_IRQ7},	/* 4 */
	{-1, 0},		/* 5 SMBUS not used */
	{-1, 0},		/* 6 */
	{-1, 0},		/* 7 */
	{-1, 0},		/* 8 */
	{PIC_INTE, PIC_IRQ5},	/* 9 */
	{PIC_INTF, PIC_IRQ5},	/* 10 */
	{PIC_INTG, PIC_IRQ5},	/* 11 */
	{PIC_INTH, PIC_IRQ4}	/* 12 */
};

enum dev_list {
	/* RS780 devices list */
	INTERNAL_GFX = 0,
	GPPSB0,
	GPPSB1,
	GPPSB2,
	GPPSB3,
	X16_SLOT,
	/* SB7xx devices list */
	HDA,
	IDE,
	SATA,
	USB1_OHCI,		// device 12h
	USB1_EHCI,		// device 12h
	USB2_OHCI,		// device 13h
	USB2_EHCI,		// device 13h
	USB3_OHCI,		// device 14h
	DEV_LIST_END
};

enum reg_int {
	REG_INTA = 0,
	REG_INTB = 1,
	REG_INTC = 2,
	REG_INTD = 3,
	REG_INTE = 4,
	REG_INTF = 5,
	REG_INTG = 6,
	REG_INTH = 7
};

struct dev_interrupt_descriptor {
	enum pic_int pic_int_number;
	enum reg_int reg_int_number;
};

struct dev_interrupt_descriptor dev_interrupt_array[DEV_LIST_END] = {
	/* INTERNAL_GFX */
	{PIC_INTC, REG_INTC},
	/* GPPSB0    */
	{PIC_INTA, REG_INTA},
	/* GPPSB1    */
	{PIC_INTB, REG_INTB},
	/* GPPSB2    */
	{PIC_INTC, REG_INTC},
	/* GPPSB3    */
	{PIC_INTD, REG_INTD},
	/* X16_SLOT  */
	{PIC_INTC, REG_INTC},
	/* HDA       */
	{PIC_INTE, REG_INTE},
	/* IDE       */
	{PIC_INTA, REG_INTA},
	/* SATA      */
	{PIC_INTH, REG_INTH},
	/* USB1_OHCI */
	{PIC_INTC, REG_INTC},
	/* USB1_EHCI */
	{PIC_INTC, REG_INTC},
	/* USB2_OHCI */
	{PIC_INTC, REG_INTC},
	/* USB2_EHCI */
	{PIC_INTC, REG_INTC},
	/* USB3_OHCI */
	{PIC_INTC, REG_INTC}
};

static int get_device_irq(enum dev_list dev)
{
	int i = 0, irq = 0;

	while (i < ARRAY_SIZE(pic_routing_array)) {
		if (pic_routing_array[i].from ==
		    dev_interrupt_array[dev].pic_int_number) {
			irq = pic_routing_array[i].to;
			break;
		}

		i++;
	}

	return irq;
}

#if 0
int __init pcibios_map_irq(struct device_t *dev, u8 slot, u8 pin)
{
	switch (dev->bus->number) {
	case 0:
	case 1:		// Integrated devices in RS780 & SB7xx
		return 0;
	case 2:
		dev->irq = get_device_irq(GPPSB0);
		pci_write_config_byte(dev, PCI_INTERRUPT_LINE, dev->irq);
		break;
	case 4:
		dev->irq = get_device_irq(GPPSB2);
		pci_write_config_byte(dev, PCI_INTERRUPT_LINE, dev->irq);
		break;
	default:		// unexcepted device
		dev->irq = 0;
		printk(KERN_ERR
		       "++++++++++++ UnExcepted Deivce! ++++++++++\n");
	}

	print_fixup_info(dev);
	return dev->irq;
}
#endif
/* 
 * SB7xx PIC can map any INTx to IRQn by programming as following:
 *
 * INTx --> 0xC00
 * IRQn --> 0xC01
 */
static inline void int2irq_map(struct pic_routing routing)
{
	outb(routing.from, 0xC00);
	outb(routing.to, 0xC01);
}

static void pic_routing_config(void)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(pic_routing_array); i++) {
		if (pic_routing_array[i].from != -1)
			int2irq_map(pic_routing_array[i]);
	}
}

static void setting_device_irq(device_t pdev, enum dev_list dev)
{
	int irq;

	irq = get_device_irq(dev);
	pci_write_config_byte(pdev, PCI_INTERRUPT_LINE, irq);
	printf("bus %x dev %x fun %x irq %x\n",
	       (pdev >> 16) & 255,
	       (pdev >> 11) & 31, (pdev >> 8) & 7, irq);
}

static void hda_initialize(device_t smbus_dev)
{
	u8 reg8;

	/* Map the HDA interrupt to INTE */
	pci_read_config_byte(smbus_dev, 0x63, &reg8);
	reg8 &= 0xf8;
	reg8 |= dev_interrupt_array[HDA].reg_int_number;
	pci_write_config_byte(smbus_dev, 0x63, reg8);

	/* Set GPIO42, GPIO43, GPIO44, GPIO46 as HD function */
	pci_write_config_word(smbus_dev, 0xf8, 0x0);
	pci_write_config_word(smbus_dev, 0xfc, 0x2 << 0);
}

static void sata_initialize(device_t smbus_dev)
{
	u8 reg8;

	/* sata interrupt map smbus reg:0Xaf map sataintmap to PCI_INTH# */
	pci_read_config_byte(smbus_dev, 0xaf, &reg8);
	reg8 &= ~(0x7 << 2);
	reg8 |= (dev_interrupt_array[SATA].reg_int_number << 2);
	pci_write_config_byte(smbus_dev, 0xaf, reg8);

	/* Set SATA and PATA Controller to combined mode
	 * Port0-Port3 is SATA mode, Port4-Port5 is IDE mode
	 */
	pci_read_config_byte(smbus_dev, 0xad, &reg8);
	reg8 |= 0x1 << 3;
	reg8 &= ~(0x1 << 4);
	pci_write_config_byte(smbus_dev, 0xad, reg8);
}

static void usb_initialize(device_t smbus_dev)
{
	u8 reg8;
	u16 reg16;

	// set USB1 & USB2 interrupt
	reg16 = (dev_interrupt_array[USB1_OHCI].reg_int_number << 0) |
	    (dev_interrupt_array[USB1_EHCI].reg_int_number << 3) |
	    (dev_interrupt_array[USB2_OHCI].reg_int_number << 8) |
	    (dev_interrupt_array[USB2_EHCI].reg_int_number << 11);
	pci_write_config_word(smbus_dev, 0xbe, reg16);

	// set USB3 interrupt
	pci_read_config_byte(smbus_dev, 0x63, &reg8);
	reg8 &= ~(0x7 << 4);
	reg8 |= (dev_interrupt_array[USB3_OHCI].reg_int_number << 4);
	pci_write_config_byte(smbus_dev, 0x63, reg8);
}

/*
 * smbus is the system control center in sb700
 */
static void loongson3a_smbus_fixup(device_t smbus_dev)
{
	usb_initialize(smbus_dev);

	sata_initialize(smbus_dev);

	hda_initialize(smbus_dev);

	pic_routing_config();

	// level or edge interrupt
	outw(inw(0x4D0) | (1 << 7) | (1 << 6) | (1 << 5) | (1 << 4) |
	     (1 << 3), 0x4D0);
}

/* fixup sb700 sata controller configure.
 *  in file "drivers/pci/quirks.c" function quirk_amd_ide_mode do the same job
 */
static void loongson3a_sata_fixup(device_t pdev)
{
	setting_device_irq(pdev, SATA);
}

static void loongson3a_ide_fixup(device_t pdev)
{
	unsigned char byte = 0;
	/*set IDE ultra DMA enable as master and slalve device */
	pci_write_config_byte(pdev, 0x54, 0xf);

	/*set ultral DAM mode 0~6  we use 6 as high speed ! */
	pci_write_config_word(pdev, 0x56,(0x6 << 0) | (0x6 << 4) | (0x6 << 8) | 
			(0x6 << 12));

	/* Set IDE native PCI-mode
	 * AMD SB710 rrg P270 [PCI_Reg : 09h]
	 * For Primary   Bit[0] : 1 Native PCI-mode, 0 Compatibility mode (Default).
	 * For Secondary Bit[2] : 1 Native PCI-mode, 0 Compatibility mode (Default).
	 */
	pci_read_config_byte(pdev, 0x9, &byte);
	byte |= (0x1 << 0) | (0x1 << 2);
	pci_write_config_word(pdev, 0x9, byte);
}

static void loongson3a_ohci1_fixup(device_t pdev)
{
	if (((pdev >> 11) & 0x1f) == 0x12)
		setting_device_irq(pdev, USB1_OHCI);
	else
		setting_device_irq(pdev, USB2_OHCI);
}

static void loongson3a_ohci2_fixup(device_t pdev)
{
	setting_device_irq(pdev, USB3_OHCI);
}

static void loongson3a_ehci_fixup(device_t pdev)
{
	if (((pdev >> 11) & 0x1f) == 0x12)
		setting_device_irq(pdev, USB1_EHCI);
	else
		setting_device_irq(pdev, USB2_EHCI);
}

static void loongson3a_lpc_fixup(device_t pdev)
{
	unsigned char t;

	pci_read_config_byte(pdev, 0x46, &t);
	pci_write_config_byte(pdev, 0x46, t | (0x3 << 6));

	pci_read_config_byte(pdev, 0x47, &t);
	pci_write_config_byte(pdev, 0x47, t | 0xff);

	pci_read_config_byte(pdev, 0x48, &t);
	pci_write_config_byte(pdev, 0x48, t | 0xff);
}

static void loongson3a_hda_fixup(device_t pdev)
{
	setting_device_irq(pdev, HDA);
}

static void loongson3a_graphic_fixup(device_t pdev)
{
	setting_device_irq(pdev, INTERNAL_GFX);
}

static void other_devices_fixup(void)
{
	int irq;
	device_t dev;
	u16 val = 0;

	// 1st r8168DL
	dev = _pci_make_tag(2, 0, 0);
	irq = get_device_irq(GPPSB0);
	pci_write_config_byte(dev, PCI_INTERRUPT_LINE, irq);

	// For PCMCIA Cardbus
	dev = _pci_make_tag(3, 0, 0);
	irq = get_device_irq(GPPSB1);
	pci_write_config_byte(dev, PCI_INTERRUPT_LINE, irq);

	// For PCMCIA mass storage
	dev = _pci_make_tag(3, 0, 1);
	irq = get_device_irq(GPPSB1);
	pci_write_config_byte(dev, PCI_INTERRUPT_LINE, irq);

	// 2nd r8168DL
	dev = _pci_make_tag(4, 0, 0);
	irq = get_device_irq(GPPSB2);
	pci_write_config_byte(dev, PCI_INTERRUPT_LINE, irq);

	// Mini pci-e solt
	dev = _pci_make_tag(5, 0, 0);
	pci_read_config_word(dev, PCI_VENDOR_ID, &val);
	if (val != 0xffff) {
		irq = get_device_irq(GPPSB3);
		pci_write_config_byte(dev, PCI_INTERRUPT_LINE, irq);
	}
}

void sb700_interrupt_fixup(void)
{
	device_t dev;

	// prepare for interrupt mapping
	dev = _pci_make_tag(0, 20, 0);
	loongson3a_smbus_fixup(dev);

	// sata interrupt setting
	dev = _pci_make_tag(0, 0x11, 0);
	loongson3a_sata_fixup(dev);

	// ide interrupt setting
	dev = _pci_make_tag(0, 0x14, 1);
	loongson3a_ide_fixup(dev);

	// USB0 interrupt setting
	dev = _pci_make_tag(0, 0x12, 0);
	loongson3a_ohci1_fixup(dev);
	dev = _pci_make_tag(0, 0x12, 1);
	loongson3a_ohci1_fixup(dev);
	dev = _pci_make_tag(0, 0x12, 2);
	loongson3a_ehci_fixup(dev);

	// USB1 interrupt setting
	dev = _pci_make_tag(0, 0x13, 0);
	loongson3a_ohci1_fixup(dev);
	dev = _pci_make_tag(0, 0x13, 1);
	loongson3a_ohci1_fixup(dev);
	dev = _pci_make_tag(0, 0x13, 2);
	loongson3a_ehci_fixup(dev);

	// USB2 interrupt setting
	dev = _pci_make_tag(0, 0x14, 5);
	loongson3a_ohci2_fixup(dev);

	// hda interrupt setting
	dev = _pci_make_tag(0, 0x14, 2);
	loongson3a_hda_fixup(dev);

	// lpc interrupt setting
	dev = _pci_make_tag(0, 0x14, 3);
	loongson3a_lpc_fixup(dev);

	// graphic interrupt setting
	dev = _pci_make_tag(1, 0x5, 0);
	loongson3a_graphic_fixup(dev);

	other_devices_fixup();
}
